// SPDX-License-Identifier: MIT
// Static resources.
// Copyright (C) 2026 Artem Senichev <artemsen@gmail.com>

#pragma once

#include "pixmap.hpp"

namespace Resource {

/** Mark icon (alpha channel mask). */
extern const Pixmap mark;

/** Image file icon (alpha channel mask). */
extern const Pixmap file;

/** Empty image icon (alpha channel mask). */
extern const Pixmap empty;

}; // namespace Resource
